#define _XOPEN_SOURCE
#define _DEFAULT_SOURCE

#include <unistd.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include "bytemess.h"
#include "shared.h"

void
usage(char *name)
{
	fprintf(stderr, "Usage: %s [-f] [-F] [-n limit] [-d datefmt] [-u unit]\n", name);
	fprintf(stderr, "Read the logbookd syslog entries\n\n");
	fprintf(stderr, "Arguments:\n");
	fprintf(stderr, "  -f          Follow the live log output\n");
	fprintf(stderr, "  -F          Same as -f but also show the old messages\n");
	fprintf(stderr, "  -n limit    Limit the amount of old messages to fetch\n");
	fprintf(stderr, "  -d datefmt  Override the date formatting on output\n");
	fprintf(stderr, "  -u unit     Filter log messages by the unit name\n");
	fprintf(stderr, "  -b boot     Boot ID specification, or 'list'\n");
	fprintf(stderr, "  -c color    Enable or disable color with 'on' and 'off'\n");
	fprintf(stderr, "  -h          Display this help text\n");
}

int
main(int argc, char *argv[])
{
	int c;
	uint32_t n = 100;
	int follow = 0;
	int history = 1;
	int color = isatty(1);
	char *datefmt = "%b %d %H:%M:%S";
	char *unit = "";
	char *end;
	long temp;
	int query_boots = 0;
	char *boot_id;
	int boot_index = 0;

	while ((c = getopt(argc, argv, "fFhn:d:u:b:c:")) != -1) {
		switch (c) {
			case 'n':
				temp = strtol(optarg, &end, 10);
				if (end == optarg) {
					fprintf(stderr, "Unable to parse -n, not a valid number\n");
					return 1;
				}
				if (temp < 0) {
					fprintf(stderr, "Negative numbers for -n are invalid\n");
					temp = 0;
				}
				n = (int) temp;
				break;
			case 'b':
				if (strcmp("list", optarg) == 0) {
					query_boots = BOOT_QUERY_LIST;
				} else {
					temp = strtol(optarg, &end, 10);
					if (end == optarg) {
						query_boots = BOOT_QUERY_ID;
						boot_id = strdup(optarg);
					} else {
						query_boots = BOOT_QUERY_INDEX;
						boot_index = abs((int) temp);
					}
				}
				break;
			case 'c':
				if (strcmp("on", optarg) == 0) {
					color = 1;
				} else if (strcmp("off", optarg) == 0) {
					color = 0;
				}
				break;
			case 'f':
				follow = 1;
				history = 0;
				break;
			case 'F':
				follow = 1;
				history = 1;
				break;
			case 'd':
				datefmt = optarg;
				break;
			case 'u':
				unit = optarg;
				break;
			case 'h':
				usage(argv[0]);
				return 0;
			case '?':
				if (optopt == 'd' || optopt == 'l') {
					fprintf(stderr, "Option -%c requires an argument.\n", optopt);
				} else if (isprint(optopt)) {
					fprintf(stderr, "Unknown option '-%c'\n", optopt);
				} else {
					fprintf(stderr, "Unknown option character x%x\n", optopt);
				}
				return 1;
			default:
				usage(argv[0]);
				return 1;
		}
	}

	int fd = socket(PF_UNIX, SOCK_SEQPACKET, 0);
	struct sockaddr_un addr;
	addr.sun_family = AF_UNIX;
	strcpy(addr.sun_path, LOGREAD_SOCKET);

	if (connect(fd, (struct sockaddr *) &addr, sizeof(addr)) == -1) {
		perror("connect");
		return 1;
	}

	unsigned char req[100];
	unsigned char *reqptr = req;

	if (query_boots == BOOT_QUERY_LIST) {
		// Query a list of boot events instead of normal logs
		reqptr = write_int8(reqptr, MODE_BOOTS);
		reqptr = write_int32(reqptr, n);
		reqptr = write_varchar(reqptr, unit);

		if (write(fd, req, reqptr - req) < 0) {
			perror("write");
			return 1;
		}
		printf("ID   BOOT ID                                FIRST ENTRY       LAST ENTRY\n");
	} else if (history == 1) {
		reqptr = write_int8(reqptr, MODE_LOG);
		reqptr = write_int32(reqptr, n);
		reqptr = write_varchar(reqptr, unit);
		reqptr = write_int8(reqptr, query_boots);
		if (query_boots == BOOT_QUERY_ID) {
			reqptr = write_varchar(reqptr, boot_id);
		} else {
			reqptr = write_int32(reqptr, boot_index);
		}

		if (write(fd, req, reqptr - req) < 0) {
			perror("write");
			return 1;
		}
	}

	unsigned char buffer[MAX_MESSAGE_SIZE];
	int boot_list_idx = 0;
	while (1) {
		ssize_t size;
		size = read(fd, buffer, MAX_MESSAGE_SIZE - 1);
		if (size < 0) {
			perror("recv");
			return 1;
		}
		if (size == 0) {
			return 0;
		}
		unsigned char *bufptr = buffer;
		uint64_t received;
		uint8_t ptype, fac, sev;
		char app[256];
		bufptr = read_int8(bufptr, &ptype);
		if (ptype == PACKET_DONE) {
			if (follow) {
				continue;
			} else {
				return 0;
			}
		}
		if (ptype == PACKET_ROW || (follow == 1 && ptype == PACKET_LIVE)) {
			bufptr = read_int64(bufptr, &received);
			bufptr = read_int8(bufptr, &sev);
			bufptr = read_int8(bufptr, &fac);
			bufptr = read_varchar(bufptr, app);
			if (strncmp(app, unit, strlen(unit)) != 0) {
				continue;
			}

			char ts[256];
			time_t stamp = (time_t) received;
			struct tm *timeinfo = localtime(&stamp);
			strftime(ts, 256, datefmt, timeinfo);
			if (color) {
				printf(GRN "[%s] " RESET BOLD "%s %s: ", ts, facility_name[fac], app);
				if (sev == 7) {
					printf(BLU);
				} else if (sev == 6) {
					printf(RESET);
				} else if (sev > 3) {
					printf(YEL);
				} else {
					printf(RED);
				}
				printf("%s\n" RESET, bufptr);
			} else {
				printf("[%s] %s %s: %s\n", ts, facility_name[fac], app, bufptr);
			}
		}
		if (ptype == PACKET_BOOT) {
			char uuid[37];
			uint64_t first;
			uint64_t last;
			bufptr = read_varchar(bufptr, uuid);
			bufptr = read_int64(bufptr, &first);
			bufptr = read_int64(bufptr, &last);

			printf("%2d   ", boot_list_idx--);
			printf("%s   ", uuid);
			char ts[256];
			time_t stamp = (time_t) first;
			struct tm *timeinfo = localtime(&stamp);
			strftime(ts, 256, datefmt, timeinfo);
			printf("%s   ", ts);
			stamp = (time_t) last;
			timeinfo = localtime(&stamp);
			strftime(ts, 256, datefmt, timeinfo);
			printf("%s\n", ts);
		}
	}
}