#include <stdlib.h>
#include <stdio.h>
#include <signal.h>

#include "greatest.h"
#include "theft.h"

#include "parser.h"

struct fuzz_env {
		size_t len_bits;
};

static enum theft_trial_res
prop_dont_crash_on_input_data(struct theft *t, void *arg1)
{
	char *input = (char *) arg1;
	size_t length = strlen(input);
	struct logmessage parsed = {0};
	parse_log_message(0, input, (ssize_t) length, &parsed);
	return THEFT_TRIAL_PASS;
}

static enum theft_trial_res
prop_message_matches(struct theft *t, void *arg1)
{
	char *input = (char *) arg1;
	size_t length = strlen(input);
	struct logmessage parsed = {0};
	parse_log_message(0, input, (ssize_t) length, &parsed);
	if (strcmp(input, parsed.msg) == 0) {
		return THEFT_TRIAL_PASS;
	} else {
		return THEFT_TRIAL_FAIL;
	}
}

static enum theft_trial_res
prop_message_rfc5424(struct theft *t, void *arg1)
{
	char *input = (char *) arg1;
	size_t length = strlen(input);
	if (length < 2) {
		return THEFT_TRIAL_SKIP;
	}
	input[0] = '1';
	input[1] = ' ';
	struct logmessage parsed = {0};
	parse_log_message(0, input, (ssize_t) length, &parsed);
	if (parsed.format != SRC_RFC5424) {
		return THEFT_TRIAL_FAIL;
	}
	if (strcmp(input, parsed.msg) == 0) {
		return THEFT_TRIAL_PASS;
	} else {
		return THEFT_TRIAL_FAIL;
	}
}

static enum theft_alloc_res
string_alloc(struct theft *t, void *penv, void **output)
{
	struct fuzz_env *env = (struct fuzz_env *) penv;

	size_t size = theft_random_bits(t, env->len_bits);
	uint8_t *buf = calloc(size + 1, sizeof(uint8_t));
	if (buf == NULL) {
		return THEFT_ALLOC_ERROR;
	}
	uint8_t num;
	for (size_t i = 0; i < size; i++) {
		num = theft_random_bits(t, 8);
		if (num == '\0') {
			num = '[';
		}
		buf[i] = num;
	}

	*output = buf;

	return THEFT_ALLOC_OK;
}

TEST test_crash_on_random_data(void)
{
	struct fuzz_env env = {
		.len_bits = 20,
	};

	static struct theft_type_info type_info = {
		.alloc = string_alloc,
		.autoshrink_config = {
			.enable = true,
		},
	};
	type_info.env = &env;

	/* Get a seed based on the current time */
	theft_seed seed = theft_seed_of_time();
	struct theft_run_config config = {
		.name = __func__,
		.prop1 = prop_dont_crash_on_input_data,
		.type_info = {&type_info},
		.seed = seed,
		.trials = 500,
		.fork = {
			.enable = true,
			.timeout = 1,
		},
		.hooks = {
			.env = (void *) &env,
		},
	};

	/* Run the property test. */
	enum theft_run_res res = theft_run(&config);
		ASSERTm("Theft success", res == THEFT_RUN_PASS);

		PASS();
}

TEST test_message_equals(void)
{
	struct fuzz_env env = {
		.len_bits = 10,
	};

	static struct theft_type_info type_info = {
		.alloc = string_alloc,
		.autoshrink_config = {
			.enable = true,
		},
	};
	type_info.env = &env;

	/* Get a seed based on the current time */
	theft_seed seed = theft_seed_of_time();
	struct theft_run_config config = {
		.name = __func__,
		.prop1 = prop_message_matches,
		.type_info = {&type_info},
		.seed = seed,
		.hooks = {
			.env = (void *) &env,
		},
	};

	/* Run the property test. */
	enum theft_run_res res = theft_run(&config);
		ASSERTm("Theft success", res == THEFT_RUN_PASS);

		PASS();
}

TEST test_rfc5424(void)
{
	struct fuzz_env env = {
		.len_bits = 10,
	};

	static struct theft_type_info type_info = {
		.alloc = string_alloc,
		.autoshrink_config = {
			.enable = true,
		},
	};
	type_info.env = &env;

	/* Get a seed based on the current time */
	theft_seed seed = theft_seed_of_time();
	struct theft_run_config config = {
		.name = __func__,
		.prop1 = prop_message_rfc5424,
		.type_info = {&type_info},
		.seed = seed,
		.hooks = {
			.env = (void *) &env,
		},
		.fork = {
			.enable = true,
			.timeout = 100,
			.signal = SIGTERM,
		},
	};

	/* Run the property test. */
	enum theft_run_res res = theft_run(&config);
		ASSERTm("Theft success", res == THEFT_RUN_PASS);

		PASS();
}

GREATEST_MAIN_DEFS();

int
main(int argc, char **argv)
{
	GREATEST_MAIN_BEGIN();
		RUN_TEST(test_crash_on_random_data);
		RUN_TEST(test_message_equals);
		RUN_TEST(test_rfc5424);

	GREATEST_MAIN_END();        /* display results */
}